#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import enum
import pymysql
from . import errors
from pymysql.err import Warning as MySQLWarning

pymysql.install_as_MySQLdb()

from sqlalchemy import create_engine, UniqueConstraint, ColumnDefault, Index
from sqlalchemy import exc as sa_exc
from sqlalchemy.sql import expression
from sqlalchemy.sql.elements import and_
from sqlalchemy import Column, DateTime, String, Integer, ForeignKey, func, Text, BigInteger, Float, UnicodeText, Enum, TIMESTAMP, FLOAT
from sqlalchemy.orm import sessionmaker, scoped_session, relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.mysql import INTEGER, LONGTEXT, BIGINT
import sqlalchemy as sa
from warnings import filterwarnings


logger = logging.getLogger(__name__)
Base = declarative_base()


class AlembicDataMigration(Base):
    """
    Alembic data migration
    """
    __tablename__ = 'alembic_version_data'
    id = Column(BigInteger, primary_key=True)
    schema_ver = Column(BigInteger)
    data_ver = Column(BigInteger)


class assign(expression.FunctionElement):
    name = 'assign'


class TransientCol(object):
    """Transient column for model projection and comparison."""
    def __init__(self, name, default=None):
        self.name = name
        self.default = default


class DbException(errors.Error):
    """Generic DB exception"""
    def __init__(self, message=None, cause=None):
        super(DbException, self).__init__(message=message, cause=cause)


class MySQL(object):
    def __init__(self, user=None, password=None, db=None, host='127.0.0.1', port=3306, *args, **kwargs):
        self.user = user
        self.password = password
        self.db = db
        self.host = host
        self.port = port

        self.engine = None
        self.session = None

        self.secure_config = None
        self.secure_query = None

    def get_connstring(self):
        con_string = 'mysql://%s:%s@%s%s/%s' % (self.user, self.password,
                                                self.host, ':%s' % self.port,
                                                self.db)
        return con_string

    def build_engine(self, connstring=None, user=None, password=None, store_as_main=True):
        try:
            filterwarnings('ignore', category=MySQLWarning)
            filterwarnings('ignore', category=sa_exc.SAWarning)

            con_str = connstring
            if con_str is None and user is not None:
                con_str = 'mysql://%s:%s@%s%s' % (user, password, self.host, ':%s' % self.port)
            if con_str is None and password is not None:
                con_str = 'mysql://%s:%s@%s%s' % ('root', password, self.host, ':%s' % self.port)
            if con_str is None:
                con_str = self.get_connstring()

            engine = create_engine(con_str, pool_size=256, max_overflow=32, pool_recycle=3600)
            if store_as_main:
                self.engine = engine

            return engine

        except Exception as e:
            logger.info('Exception in building MySQL DB engine %s' % e)
            raise

    def init_db(self):
        self.build_engine()
        self.session = scoped_session(sessionmaker(bind=self.engine))

        # Make sure tables are created
        Base.metadata.create_all(self.engine)

    def get_session(self):
        return self.session()

    def get_engine(self):
        return self.engine

    def execute_sql(self, sql=None, engine=None, ignore_fail=False):
        try:
            if engine is None:
                engine = self.engine

            res = engine.execute(sql)
            return res

        except Exception as e:
            logger.debug('Exception in sql: %s, %s' % (sql, e))
            if not ignore_fail:
                raise

        finally:
            pass

        return None


#
# Helper functions
#


def silent_close(c, quiet=True):
    # noinspection PyBroadException
    try:
        if c is not None:
            c.close()
    except Exception as e:
        if not quiet:
            logger.error('Close exception: %s' % e)


def silent_rollback(c, quiet=True):
    # noinspection PyBroadException
    try:
        if c is not None:
            c.rollback()
    except Exception as e:
        if not quiet:
            logger.error('Rollback exception: %s' % e)


def silent_expunge_all(c, quiet=True):
    # noinspection PyBroadException
    try:
        if c is not None:
            c.expunge_all()()
    except Exception as e:
        if not quiet:
            logger.error('Expunge exception: %s' % e)


#
# DB Entities
#


class StatusEnum(enum.Enum):
    pending = 'pending'
    running = 'running'
    finished = 'finished'
    error = 'error'


class TestResultEnum(enum.Enum):
    passed = 'passed'
    failed = 'failed'


class WorkerTypeEnum(enum.Enum):
    longterm = 'longterm'
    shortterm = 'shortterm'


class Experiments(Base):
    __tablename__ = 'experiments'
    id = Column(BigInteger, primary_key=True)
    name = Column(String(255), nullable=False)
    author_email = Column(String(255), nullable=True)
    status = Column(Enum(StatusEnum), default=StatusEnum.pending)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    run_started = Column(DateTime, nullable=True)
    run_finished = Column(DateTime, nullable=True)
    config_file = Column(String(255), nullable=False)
    data_file = Column(String(255), nullable=False)
    data_file_sha256 = Column(String(64), nullable=False)


class Jobs(Base):
    __tablename__ = 'jobs'
    id = Column(BigInteger, primary_key=True)
    battery = Column(String(100), nullable=False)
    status = Column(Enum(StatusEnum), default=StatusEnum.pending.value)  #, key=True)  # enum key is problem
    run_started = Column(DateTime, nullable=True)
    run_finished = Column(DateTime, nullable=True)
    experiment_id = Column(ForeignKey('experiments.id', name='jobs_experiments_id', ondelete='CASCADE'),
                           nullable=False, index=True, primary_key=False)
    run_heartbeat = Column(DateTime, nullable=True)
    worker_id = Column(BigInteger, nullable=True)
    worker_pid = Column(Integer, nullable=True)
    retries = Column(Integer, nullable=False, default=0)
    lock_version = Column(Integer, nullable=False, default=0)


class Batteries(Base):
    __tablename__ = 'batteries'
    id = Column(BigInteger, primary_key=True)
    name = Column(String(100), nullable=False)
    passed_tests = Column(BigInteger, nullable=False)
    total_tests = Column(BigInteger, nullable=False)
    alpha = Column(Float, nullable=False)
    experiment_id = Column(ForeignKey('experiments.id', name='batteries_experiments_id', ondelete='CASCADE'),
                           nullable=False, index=True, primary_key=False)
    job_id = Column(ForeignKey('jobs.id', name='batteries_jobs__id', ondelete='CASCADE'),
                    nullable=False, index=True, primary_key=False)
    pvalue = Column(Float, nullable=True)
    experiment = relationship('Experiments', foreign_keys=experiment_id)
    job = relationship('Jobs', foreign_keys=job_id)


class BatteryErrors(Base):
    __tablename__ = 'battery_errors'
    id = Column(BigInteger, primary_key=True)
    message = Column(Text, nullable=False)
    battery_id = Column(ForeignKey('batteries.id', name='battery_errors_batteries_id', ondelete='CASCADE'),
                        nullable=False, index=True, primary_key=False)
    battery = relationship('Batteries', foreign_keys=battery_id)


class BatteryWarnings(Base):
    __tablename__ = 'battery_warnings'
    id = Column(BigInteger, primary_key=True)
    message = Column(Text, nullable=False)
    battery_id = Column(ForeignKey('batteries.id', name='battery_warnings_batteries_id', ondelete='CASCADE'),
                        nullable=False, index=True, primary_key=False)
    battery = relationship('Batteries', foreign_keys=battery_id)


class Tests(Base):
    __tablename__ = 'tests'
    id = Column(BigInteger, primary_key=True)
    name = Column(String(255), nullable=False)
    partial_alpha = Column(Float, nullable=False)
    result = Column(Enum(TestResultEnum), nullable=False)
    test_index = Column(Integer, nullable=False)
    battery_id = Column(ForeignKey('batteries.id', name='tests_batteries_id', ondelete='CASCADE'),
                        nullable=False, index=True, primary_key=False)
    pvalue = Column(Float, nullable=True)
    battery = relationship('Batteries', foreign_keys=battery_id)


class Variants(Base):
    __tablename__ = 'variants'
    id = Column(BIGINT(unsigned=True), primary_key=True)
    variant_index = Column(Integer, nullable=False)
    test_id = Column(ForeignKey('tests.id', name='variants_tests_id', ondelete='CASCADE'),
                     nullable=False, index=True, primary_key=False)
    test = relationship('Tests', foreign_keys=test_id)


class VariantsErrors(Base):
    __tablename__ = 'variant_errors'
    id = Column(BigInteger, primary_key=True)
    message = Column(Text, nullable=False)
    variant_id = Column(ForeignKey('variants.id', name='variant_errors_variants_id', ondelete='CASCADE'),
                        nullable=False, index=True, primary_key=False)
    variant = relationship('Variants', foreign_keys=variant_id)


class VariantWarnings(Base):
    __tablename__ = 'variant_warnings'
    id = Column(BigInteger, primary_key=True)
    message = Column(Text, nullable=False)
    variant_id = Column(ForeignKey('variants.id', name='variant_warnings_variants_id', ondelete='CASCADE'),
                        nullable=False, index=True, primary_key=False)
    variant = relationship('Variants', foreign_keys=variant_id)


class VariantStderr(Base):
    __tablename__ = 'variant_stderr'
    id = Column(BigInteger, primary_key=True)
    message = Column(Text, nullable=False)
    variant_id = Column(ForeignKey('variants.id', name='variant_stderr_variants_id', ondelete='CASCADE'),
                        nullable=False, index=True, primary_key=False)
    variant = relationship('Variants', foreign_keys=variant_id)


class VariantResults(Base):
    __tablename__ = 'variant_results'
    id = Column(BigInteger, primary_key=True)
    message = Column(Text, nullable=False)
    variant_id = Column(ForeignKey('variants.id', name='variant_results_variants_id', ondelete='CASCADE'),
                        nullable=False, index=True, primary_key=False)
    variant = relationship('Variants', foreign_keys=variant_id)


class UserSettings(Base):
    __tablename__ = 'user_settings'
    id = Column(BigInteger, primary_key=True)
    name = Column(String(100), nullable=False)
    value = Column(String(50), nullable=False)
    variant_id = Column(ForeignKey('variants.id', name='user_settings_variants_id', ondelete='CASCADE'),
                        nullable=False, index=True, primary_key=False)
    variant = relationship('Variants', foreign_keys=variant_id)


class RttSettings(Base):
    __tablename__ = 'rtt_settings'
    __table_args__ = (UniqueConstraint('name', name='rtt_settings_name_unique'),)
    id = Column(BigInteger, primary_key=True)
    name = Column(String(100), nullable=False)
    value = Column(String(50), nullable=False)


class Subtests(Base):
    __tablename__ = 'subtests'
    id = Column(BigInteger, primary_key=True)
    subtest_index = Column(Integer, nullable=False)
    variant_id = Column(ForeignKey('variants.id', name='subtests_variants_id', ondelete='CASCADE'),
                        nullable=False, index=True, primary_key=False)
    variant = relationship('Variants', foreign_keys=variant_id)


class Statistics(Base):
    __tablename__ = 'statistics'
    id = Column(BigInteger, primary_key=True)
    name = Column(String(255), nullable=False)
    value = Column(Float, nullable=False)
    result = Column(Enum(TestResultEnum), nullable=False)
    subtest_id = Column(ForeignKey('subtests.id', name='statistics_subtests_id', ondelete='CASCADE'),
                        nullable=False, index=True, primary_key=False)
    subtest = relationship('Subtests', foreign_keys=subtest_id)


class TestParameters(Base):
    __tablename__ = 'test_parameters'
    id = Column(BigInteger, primary_key=True)
    name = Column(String(100), nullable=False)
    value = Column(String(50), nullable=False)
    subtest_id = Column(ForeignKey('subtests.id', name='test_parameters_subtests_id', ondelete='CASCADE'),
                        nullable=False, index=True, primary_key=False)
    subtest = relationship('Subtests', foreign_keys=subtest_id)


class Pvalues(Base):
    __tablename__ = 'p_values'
    id = Column(BigInteger, primary_key=True)
    value = Column(Float, nullable=False)
    subtest_id = Column(ForeignKey('subtests.id', name='p_values_subtests_id', ondelete='CASCADE'),
                        nullable=False, index=True, primary_key=False)
    subtest = relationship('Subtests', foreign_keys=subtest_id)


class Workers(Base):
    __tablename__ = 'workers'
    __table_args__ = (UniqueConstraint('worker_id', name='workers_worker_id_unique'),)
    id = Column(BigInteger, primary_key=True)
    worker_id = Column(String(32), nullable=False)
    worker_name = Column(String(250), nullable=True)
    worker_type = Column(Enum(WorkerTypeEnum), default=WorkerTypeEnum.longterm)
    worker_added = Column(DateTime, nullable=True)
    worker_last_seen = Column(DateTime, nullable=True)
    worker_active = Column(Integer, nullable=False, default=1)
    worker_address = Column(String(250), nullable=True)
    worker_location = Column(String(250), nullable=True)
    worker_aux = Column(Text, nullable=True)









#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import coloredlogs
import logging
import json
import itertools
import shlex
import time
import queue
import sys
from jsonpath_ng import jsonpath, parse

from .database import MySQL, Jobs, Experiments, Batteries, Variants, Tests, \
    TestResultEnum, VariantStderr, VariantResults, BatteryErrors, \
    Subtests, Statistics, TestParameters, Pvalues, UserSettings, \
    silent_close, silent_expunge_all, silent_rollback
from booltest.runner import AsyncRunner
from .utils import merge_pvals, booltest_pval


logger = logging.getLogger(__name__)
coloredlogs.install(level=logging.INFO)


def jsonpath(path, obj, allow_none=False):
    r = [m.value for m in parse(path).find(obj)]
    return r[0] if not allow_none else (r[0] if r else None)


def listize(obj):
    return obj if (obj is None or isinstance(obj, list)) else [obj]


def get_runner(cli, cwd=None, rtt_env=None):
    async_runner = AsyncRunner(cli, cwd=cwd, shell=False, env=rtt_env)
    async_runner.log_out_after = False
    async_runner.preexec_setgrp = True
    return async_runner


class BoolParamGen:
    def __init__(self, cli, vals):
        self.cli = cli
        self.vals = vals if isinstance(vals, list) else [vals]


class BoolJob:
    def __init__(self, cli, name, vinfo='', idx=None):
        self.cli = cli
        self.name = name
        self.vinfo = vinfo
        self.idx = idx

    def is_halving(self):
        return '--halving' in self.cli


class BoolRes:
    def __init__(self, job, ret_code, js_res, is_halving, rejects=False, pval=None, alpha=None, stderr=None):
        self.job = job  # type: BoolJob
        self.ret_code = ret_code
        self.js_res = js_res
        self.is_halving = is_halving
        self.rejects = rejects
        self.alpha = alpha
        self.pval = pval
        self.stderr = stderr


class BoolRunner:
    def __init__(self):
        self.args = None
        self.db = None
        self.rtt_config = None
        self.bool_config = None
        self.parallel_tasks = None
        self.bool_wrapper = None
        self.job_queue = queue.Queue(maxsize=0)
        self.runners = []  # type: List[Optional[AsyncRunner]]
        self.comp_jobs = []  # type: List[Optional[BoolJob]]
        self.results = []

    def init_config(self):
        self.parallel_tasks = self.args.threads or 1
        try:
            with open(self.args.rtt_config) as fh:
                self.rtt_config = json.load(fh)

            self.bool_config = jsonpath('"toolkit-settings"."booltest"', self.rtt_config, False)
            if not self.bool_wrapper:
                self.bool_wrapper = jsonpath("$.wrapper", self.bool_config, True)
            if not self.args.threads:
                self.parallel_tasks = jsonpath('$."toolkit-settings".execution."max-parallel-tests"', self.rtt_config, True) or 1

        except Exception as e:
            logger.error("Could not load RTT config %s" % (e,), exc_info=e)

        if not self.bool_wrapper:
            self.bool_wrapper = "\"%s\" -m booltest.booltest_main" % sys.executable

    def init_db(self):
        if self.args.no_db:
            return
        if self.rtt_config is None:
            logger.debug("Could not init DB, no config given")
            return

        db_cfg = [m.value for m in parse('"toolkit-settings"."result-storage"."mysql-db"').find(self.rtt_config)][0]
        db_creds = self.args.db_creds if self.args.db_creds else db_cfg["credentials-file"]
        with open(db_creds) as fh:
            creds = json.load(fh)

        uname = [m.value for m in parse('"credentials"."username"').find(creds)][0]
        passwd = [m.value for m in parse('"credentials"."password"').find(creds)][0]
        host = self.args.db_host if self.args.db_host else db_cfg['address']
        port = self.args.db_port if self.args.db_port else int(db_cfg['port'])

        try:
            self.db = MySQL(user=uname, password=passwd, db=db_cfg['name'], host=host, port=port)
            self.db.init_db()
        except Exception as e:
            logger.warning("Exception in DB connect %s" % (e,), exc_info=e)
            self.db = None

    def generate_jobs(self):
        dcli = jsonpath('$.default-cli', self.bool_config, True) or ''
        strategies = jsonpath('$.strategies', self.bool_config, False)

        for st in strategies:
            name = st['name']
            st_cli = jsonpath('$.cli', st, True) or ''
            st_vars = jsonpath('$.variations', st, True) or []
            ccli = ('%s %s' % (dcli, st_cli)).strip()

            if not st_vars:
                yield BoolJob(ccli, name)
                continue

            for cvar in st_vars:
                blocks = listize(jsonpath('$.bl', cvar, True)) or [None]
                degs = listize(jsonpath('$.deg', cvar, True)) or [None]
                cdegs = listize(jsonpath('$.cdeg', cvar, True)) or [None]
                pcli = ['--block', '--degree', '--combine-deg']
                vinfo = ['', '', '']
                iterator = itertools.product(blocks, degs, cdegs)

                for el in iterator:
                    c = ' '.join([(('%s %s') % (pcli[ix], dt)) for (ix, dt) in enumerate(el) if dt is not None])
                    vi = '-'.join([(('%s%s') % (vinfo[ix], dt)).strip() for (ix, dt) in enumerate(el) if dt is not None])
                    ccli0 = ('%s %s' % (ccli, c)).strip()

                    yield BoolJob(ccli0, name, vi)

    def run_job(self, cli):
        async_runner = get_runner(shlex.split(cli))

        logger.info("Starting async command %s" % cli)
        async_runner.start()

        while async_runner.is_running:
            time.sleep(1)
        logger.info("Async command finished")

    def on_finished(self, job, runner, idx):
        if runner.ret_code != 0:
            logger.warning("Return code of job %s is %s" % (idx, runner.ret_code))
            stderr = ("\n".join(runner.err_acc)).strip()
            br = BoolRes(job, runner.ret_code, None, job.is_halving, stderr=stderr)
            self.results.append(br)
            return

        results = runner.out_acc
        buff = (''.join(results)).strip()
        try:
            js = json.loads(buff)

            is_halving = js['halving']
            br = BoolRes(job, 0, js, is_halving)

            if not is_halving:
                br.rejects = [m.value for m in parse('$.inputs[0].res[0].rejects').find(js)][0]
                br.alpha = [m.value for m in parse('$.inputs[0].res[0].ref_alpha').find(js)][0]
                logger.info('rejects: %s, at alpha %.5e' % (br.rejects, br.alpha))

            else:
                br.pval = [m.value for m in parse('$.inputs[0].res[1].halvings[0].pval').find(js)][0]
                logger.info('halving pval:', br.pval)

            self.results.append(br)

        except Exception as e:
            logger.error("Exception processing results: %s" % (e,), exc_info=e)
            logger.info("[[[%s]]]" % buff)

    def on_results_ready(self):
        if self.args.no_db or self.args.eid < 0 or self.args.jid < 0:
            logger.info("Results will not be inserted to the database. ")
            return

        s = None
        try:
            s = self.db.get_session()
            job_db = s.query(Jobs).filter(Jobs.id == self.args.jid).first()
            exp_db = s.query(Experiments).filter(Experiments.id == self.args.eid).first()

            if not job_db:
                logger.info("Results store fail, could not load Job with id %s" % (self.args.jid,))
                return

            if not exp_db:
                logger.info("Results store fail, could not load Experiment with id %s" % (self.args.eid,))
                return

            bat_db = Batteries(name=self.args.battery, passed_tests=0, total_tests=1, alpha=self.args.alpha,
                               experiment_id=self.args.eid, job_id=self.args.jid)
            s.add(bat_db)
            s.flush()

            bat_errors = ['Job %d (%s-%s), ret_code %d' % (r.job.idx, r.job.name, r.job.vinfo, r.ret_code)
                          for r in self.results if r.ret_code != 0]
            if bat_errors:
                bat_err_db = BatteryErrors(message='\n'.join(bat_errors), battery=bat_db)
                s.add(bat_err_db)

            ok_results = [r for r in self.results if r.ret_code == 0]
            pvalue = -1
            if self.is_halving_battery():
                pvals = [r.pval for r in ok_results]
                npassed = sum([1 for r in ok_results if r.pval >= self.args.alpha])
                pvalue = merge_pvals(pvals)[0] if len(pvals) > 1 else -1

            else:
                rejects = [r for r in ok_results if r.rejects]
                alpha = max([x.alpha for x in ok_results])
                pvalue = booltest_pval(nfails=len(rejects), ntests=len(ok_results), alpha=alpha)
                npassed = sum([1 for r in ok_results if not r.rejects])

            bat_db.total_tests = len(ok_results)
            bat_db.passed_tests = npassed
            bat_db.pvalue = pvalue
            bat_db = s.merge(bat_db)

            for rs in self.results:  # type: BoolRes
                passed = (rs.pval >= self.args.alpha if rs.is_halving else not rs.rejects) if rs.ret_code == 0 else None
                passed_res = (TestResultEnum.passed if passed else TestResultEnum.failed) if passed is not None else TestResultEnum.passed

                test_db = Tests(name="%s %s" % (rs.job.name, rs.job.vinfo), partial_alpha=self.args.alpha,
                                result=passed_res, test_index=rs.job.idx, battery=bat_db)
                s.add(test_db)

                var_db = Variants(variant_index=0, test=test_db)
                s.add(var_db)

                uset_db = UserSettings(name="Cfg", value=rs.job.vinfo, variant=var_db)
                s.add(uset_db)

                if rs.ret_code != 0:
                    var_err_db = VariantStderr(message=rs.stderr, variant=var_db)
                    s.add(var_err_db)
                    continue

                var_res_db = VariantResults(message=json.dumps(rs.js_res), variant=var_db)
                s.add(var_res_db)

                sub_db = Subtests(subtest_index=0, variant=var_db)
                s.add(sub_db)

                if rs.is_halving:
                    st_db = Statistics(name="pvalue", value=rs.pval, result=passed_res, subtest=sub_db)
                    pv_db = Pvalues(value=rs.pval, subtest=sub_db)
                    test_db.pvalue = rs.pval
                    s.add(st_db)
                    s.add(pv_db)

                else:
                    cpval = rs.alpha - 1e-20 if rs.rejects else 1
                    st_db = Statistics(name="pvalue", value=cpval, result=passed_res, subtest=sub_db)
                    tp_db = TestParameters(name="alpha", value=rs.alpha, subtest=sub_db)
                    test_db.pvalue = cpval
                    s.add(st_db)
                    s.add(tp_db)

                s.merge(test_db)
            s.commit()

        except Exception as e:
            logger.warning("Exception in storing results: %s" % (e,), exc_info=e)

        finally:
            silent_expunge_all(s)
            silent_close(s)

    def is_halving_battery(self):
        return self.args.battery == 'booltest_2'

    def work(self):
        jobs = [x for x in self.generate_jobs() if x.is_halving() == (self.is_halving_battery())]
        for i, j in enumerate(jobs):
            j.idx = i

        self.runners = [None] * self.parallel_tasks
        self.comp_jobs = [None] * self.parallel_tasks

        for j in jobs:
            self.job_queue.put_nowait(j)

        while not self.job_queue.empty() or sum([1 for x in self.runners if x is not None]) > 0:
            time.sleep(0.1)

            # Realloc work
            for i in range(len(self.runners)):
                if self.runners[i] is not None and self.runners[i].is_running:
                    continue

                was_empty = self.runners[i] is None
                if not was_empty:
                    self.job_queue.task_done()
                    logger.info("Task %d done, job queue size: %d, running: %s"
                                % (i, self.job_queue.qsize(), sum([1 for x in self.runners if x])))
                    self.on_finished(self.comp_jobs[i], self.runners[i], i)

                # Start a new task, if any
                try:
                    job = self.job_queue.get_nowait()  # type: BoolJob
                except queue.Empty:
                    self.runners[i] = None
                    continue

                cli = '%s %s "%s"' % (self.bool_wrapper, job.cli, self.args.data_path)
                self.comp_jobs[i] = job
                self.runners[i] = get_runner(shlex.split(cli))
                logger.info("Starting async command %s %s, %s" % (job.name, job.vinfo, cli))
                self.runners[i].start()

        self.on_results_ready()

    def main(self):
        logger.debug('App started')

        parser = self.argparser()
        self.args = parser.parse_args()
        self.init_config()
        self.init_db()
        self.work()

    def argparser(self):
        parser = argparse.ArgumentParser(description='BoolTest RTT runner')

        parser.add_argument('--debug', dest='debug', action='store_const', const=True,
                            help='enables debug mode')
        parser.add_argument('-s', '--rtt-config', dest='rtt_config',
                            help='RTT Configuration path')
        parser.add_argument('-b', '--battery', default=None,
                            help='Battery to execute')
        parser.add_argument('-c', '--config', default=None,
                            help='Job config')
        parser.add_argument('-f', '--data-path', dest='data_path', default=None,
                            help='Job data path')
        parser.add_argument('--eid', type=int, default=-1,
                            help='Experiment ID')
        parser.add_argument('--jid', type=int, default=-1,
                            help='Job ID')
        parser.add_argument('--db-host', dest='db_host',
                            help='MySQL host name')
        parser.add_argument('--db-port', dest='db_port', type=int, default=None,
                            help='MySQL port')
        parser.add_argument('--db-creds', dest='db_creds',
                            help='MySQL credentials json')
        parser.add_argument('--rpath',
                            help='Experiment dir')
        parser.add_argument('--no-db', dest='no_db', action='store_const', const=True,
                            help='No database connection')
        parser.add_argument('--alpha', dest='alpha', type=float, default=1e-4,
                            help='Alpha value for pass/fail')
        parser.add_argument('-t', dest='threads', type=int, default=None,
                            help='Maximum parallel threads')
        return parser


def main():
    br = BoolRunner()
    return br.main()


if __name__ == '__main__':
    main()



import scipy.misc
import math

if hasattr(scipy.misc, 'comb'):
    scipy_comb = scipy.misc.comb
else:
    import scipy.special
    scipy_comb = scipy.special.comb


def chunks(items, size):
    for i in range(0, len(items), size):
        yield items[i : i + size]


def sidak_alpha(alpha, m):
    """
    Compute new significance level alpha for M independent tests.
    More tests -> unexpected events occur more often thus alpha has to be adjusted.
    Overall test battery fails if min(pvals) < new_alpha.
    """
    return 1 - (1 - alpha)**(1./m)


def sidak_inv(alpha, m):
    """
    Inverse transformation of sidak_alpha function.
    Used to compute final p-value of M independent tests if while preserving the
    same significance level for the resulting p-value.
    """
    return 1 - (1 - alpha)**m


def merge_pvals(pvals, batch=2):
    """
    Merging pvals with Sidak.

    Note that the merging tree has to be symmetric, otherwise the computation on pvalues is not correct.
    Note: 1-(1-(1-(1-x)^3))^2 == 1-((1-x)^3)^2 == 1-(1-x)^6.
    Example: 12 nodes, binary tree: [12] -> [2,2,2,2,2,2] -> [2,2,2]. So far it is symmetric.
    The next layer of merge is problematic as we merge [2,2] and [2] to two p-values.
    If a minimum is from [2,2] (L) it is a different expression as from [2] R as the lists
    have different lengths. P-value from [2] would increase in significance level compared to Ls on this new layer
    and this it has to be corrected.
    On the other hand, the L minimum has to be corrected as well as it came from
    list of the length 3. We want to obtain 2 p-values which can be merged as if they were equal (exponent 2).
    Thus the exponent on the [2,2] and [2] layer will be 3/2 as we had 3 p-values in total and we are producing 2.
    """
    if len(pvals) <= 1:
        return pvals

    batch = min(max(2, batch), len(pvals))  # norm batch size
    parts = list(chunks(pvals, batch))
    exponent = len(pvals) / len(parts)
    npvals = []
    for p in parts:
        pi = sidak_inv(min(p), exponent)
        npvals.append(pi)
    return merge_pvals(npvals, batch)


def booltest_pval(nfails=1, ntests=36, alpha=1/20000):
    acc = [scipy_comb(ntests, k) * (1 - alpha) ** (ntests - k) * alpha ** k for k in range(nfails)]
    return max(0, 1 - sum(acc))


def booltest_pval_log(nfails=1, ntests=36, alpha=1/20000):
    log = math.log2
    acc = [2 ** (log(scipy_comb(ntests, k)) + (ntests - k) * log((1 - alpha)) + k * log(alpha)) for k in range(nfails)]
    return max(0, 1 - sum(acc))



import sys

from setuptools import setup
from setuptools import find_packages

version = '0.6.3'

# Please update tox.ini when modifying dependency version requirements
install_requires = [
    'booltest',
    'setuptools>=1.0',
    'six',
    'future',
    'requests',
    'coloredlogs',
    'jsonpath-ng',

    'sarge>=0.1.4',
    'SQLAlchemy>=1.1',
    'SqlAlchemy-Enum-Tables',
    'pymysql',
    'shellescape',
    'scipy',

]

dev_extras = [
    'nose',
    'pep8',
    'tox',
    'pypandoc',
]

docs_extras = [
    'Sphinx>=1.0',  # autodoc_member_order = 'bysource', autodoc_default_flags
    'sphinx_rtd_theme',
    'sphinxcontrib-programoutput',
]

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
    long_description = long_description.replace("\r", '')

except(IOError, ImportError):
    import io
    with io.open('README.md', encoding="utf-8") as f:
        long_description = f.read()

setup(
    name='booltest-rtt',
    version=version,
    description='BoolTest: RTT wrapper',
    long_description=long_description,
    url='https://github.com/ph4r05/booltest-rtt',
    author='Dusan Klinec',
    author_email='dusan.klinec@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Security',
    ],

    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        'dev': dev_extras,
        'docs': docs_extras,
    },
    entry_points={
        'console_scripts': [
            'booltest_rtt = booltest_rtt.main:main',
        ],
    }
)
